#!/usr/bin/env python3
import uuid

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Corpus


class Command(BaseCommand):
    help = "Reindex elements, transcriptions and classifications into Solr"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--all", help="Reindex all corpus", action="store_true",
        )
        parser.add_argument(
            "-c", "--corpus-id", help="UUID of an existing corpus to reindex", type=uuid.UUID
        )
        parser.add_argument(
            "--drop",
            help="Drop the existing indexes before reindexing",
            action="store_true",
        )
        parser.add_argument(
            "--setup",
            help="Only setup a collection. Create a collection and fields if they do not exist or update the fields",
            action="store_true",
        )
        parser.add_argument(
            "--sql-chunk-size",
            help="Size of the chunks of parent elements used when retrieving all children of indexable elements from the database.",
            type=int,
            default=10000,
        )

    def handle(self, corpus_id, sql_chunk_size, **options):
        if not settings.ARKINDEX_FEATURES["search"]:
            raise CommandError("Reindexation is not possible if the search feature flag is disabled. "
                               "Consider setting `features.search` to `on` or `true` or `yes` in the YAML "
                               "configuration file, and configuring Solr properly.")

        if sql_chunk_size <= 0:
            raise CommandError("--sql-chunk-size must be set to a strictly positive integer.")

        if options["all"]:
            corpora = Corpus.objects.filter(indexable=True)
        elif corpus_id:
            corpora = Corpus.objects.filter(id=corpus_id, indexable=True)
            if not corpora.exists():
                raise CommandError(f"Corpus {corpus_id} does not exist or is not indexable")
        else:
            raise CommandError("Either specify a corpus ID with --corpus-id or index all corpus through --all")

        for corpus in corpora:
            self.stdout.write(f"Indexing {corpus.name}")
            indexer = Indexer(corpus.id, sql_chunk_size=sql_chunk_size)
            if options.get("drop"):
                indexer.drop_index()
            indexer.setup()
            if options.get("setup"):
                continue
            indexer.index()
