from datetime import datetime, timedelta, timezone
from unittest.mock import call, patch
from uuid import uuid4

from django.test import override_settings
from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.documents.models import Corpus, CorpusExport, CorpusExportState
from arkindex.ponos.models import Farm
from arkindex.process.models import ArkindexFeature, Process, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User

DB_CREATED = "2020-02-02T01:23:45.678000Z"


@override_settings(EXPORT_SOURCES=["default", "cached_db"], PONOS_MAXIMUM_TASK_TTL=3600)
class TestExportProcess(FixtureAPITestCase):
    """
    Test the creation of export processes
    """
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = DB_CREATED
            cls.complete_export = CorpusExport.objects.create(corpus=cls.corpus, user=cls.user, state=CorpusExportState.Done, source="default")
        cls.other_corpus = Corpus.objects.create(name="Das korpus der anderen", public=True)
        cls.corpus_element = cls.corpus.elements.first()
        cls.other_element = cls.other_corpus.elements.create(
            type=cls.other_corpus.types.create(slug="a_type", display_name="an element type"),
            name="alien element",
        )
        cls.pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        cls.pdf_export_version.configuration = {
            "user_configuration": {
                "page_type": {"title": "Page element type", "type": "string", "default": "page", "required": False},
                "order_by_name": {"title": "Order by name or nah", "type": "bool", "required": False, "default": False},
                "example_int": {"title": "Some integer parameter", "type": "int", "required": False, "default": 2},
                "example_choice": {"title": "Some choice parameter", "type": "enum", "required": False, "choices": ["a", "b", "c"], "default": "a"},
                "example_required": {"title": "Some parameter that is required and has no default", "type": "string", "required": True},
                "example_list": {"title": "Some list parameter", "type": "list", "subtype": "string", "default": ["pomme", "poire"], "required": False},
                "example_list_2": {"title": "Some other list parameter", "type": "list", "subtype": "string", "default": ["prune", "abricot"], "required": False}
            }
        }
        cls.pdf_export_version.save()

    def test_requires_logged_in(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_requires_verified(self):
        test_user = User.objects.create(email="kaji@seele.co.jp")
        self.client.force_login(test_user)

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_export_id_required(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.other_corpus.id)}),
                {
                    "format": "pdf"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"export_id": ["This field is required."]})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_no_corpus_guest_access(self, filter_rights_mock):
        test_user = User.objects.create(email="katsuragi@nerv.co.jp", verified_email=True)
        self.client.force_login(test_user)

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})
        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(test_user, Corpus, Role.Guest.value))


    @patch("arkindex.users.utils.get_max_level")
    def test_requires_corpus_admin(self, max_level_mock):
        for lvl in [Role.Guest.value, Role.Contributor.value]:
            max_level_mock.return_value = lvl
            self.client.force_login(self.user)

            with self.assertNumQueries(3):
                response = self.client.post(
                    reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                    {
                        "format": "pdf",
                        "export_id": None,
                    },
                )
                self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
            self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to this corpus."})
        self.assertEqual(max_level_mock.call_count, 2)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.corpus))


    def test_export_in_corpus(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.other_corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"export_id": [f"Export {str(self.complete_export.id)} does not belong to corpus {str(self.other_corpus.id)}."]})

    def test_failed_export(self):
        failed_export = CorpusExport.objects.create(corpus=self.corpus, user=self.user, state=CorpusExportState.Failed, source="default")
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(failed_export.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"export_id": [f"Export {str(failed_export.id)} has failed."]})

    def test_export_already_running(self):
        CorpusExport.objects.create(corpus=self.corpus, user=self.user, state=CorpusExportState.Running, source="default")
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"export_id": ["An export is already running for this corpus."]})

    @override_settings(EXPORT_TTL_SECONDS=420)
    def test_recent_export_exists(self):
        self.client.force_login(self.user)
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(minutes=2)
            CorpusExport.objects.create(corpus=self.corpus, user=self.user, state=CorpusExportState.Done, source="default")

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"export_id": ["An export has already been made for this corpus in the last 420 seconds."]})

    def test_element_does_not_exist(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "element_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"element_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.']})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_element_non_readable_corpus(self, filter_rights_mock):
        filter_rights_mock.return_value = Corpus.objects.filter(pk=self.corpus.id)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "element_id": str(self.other_element.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"element_id": [f'Invalid pk "{str(self.other_element.id)}" - object does not exist.']})

    def test_element_wrong_corpus(self):
        self.other_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "element_id": str(self.other_element.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"element_id": [f"Element {str(self.other_element.id)} does not belong to corpus {str(self.corpus.id)}."]})

    def test_unsupported_format(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "chaussette",
                    "export_id": str(self.complete_export.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"format": ["Value is not of type ExportFormat"]})

    def test_element_or_selection(self):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.corpus_element)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "element_id": str(self.corpus_element.id),
                    "selection": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "non_field_errors": ["Only one of `element` and `selection` may be set."]
        })

    def test_empty_selection(self):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.other_element)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "selection": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"selection": [f"The user selection does not contain any elements from corpus {str(self.corpus.id)}."]})

    def test_bad_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "configuration": {
                        "page_type": None,
                        "order_by_name": "nope",
                        "example_int": "a",
                        "example_choice": "e",
                        "example_list": "abricot",
                        "example_list_2": [12, "prune"],
                        "bad_param": "hello"
                    }
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "configuration": {
                "bad_param": ["This is a not a valid parameter for this worker version."],
                "example_required": ["This parameter is required."],
                "example_choice": ["e is not an available choice."],
                "order_by_name": ["This is not a valid value for a field of type bool."],
                "example_int": ["This is not a valid value for a field of type int."],
                "example_list": ["This is not a valid value for a field of type list."],
                "example_list_2": ["All items must be of type str."],
                "page_type": ["This is not a valid value for a field of type string."]
            }
        })

    def test_create_export_process(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(24):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "configuration": {
                        "example_required": "something"
                    }
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_process = Process.objects.get(mode=ProcessMode.Export)
        created_task = created_process.tasks.get()
        created_run = created_process.worker_runs.get()

        self.assertDictEqual(created_run.configuration.configuration, {
            "example_choice": "a",
            "example_int": 2,
            "example_list": ["pomme", "poire"],
            "example_list_2": ["prune", "abricot"],
            "example_required": "something",
            "export_id": str(self.complete_export.id),
            "order_by_name": False,
            "page_type": "page"
        })

        self.assertDictEqual(response.json(), {
            "name": None,
            "mode": "export",
            "id": str(created_process.id),
            "state": "unscheduled",
            "corpus": str(self.corpus.id),
            "created": created_process.created.isoformat().replace("+00:00", "Z"),
            "started": created_process.started.isoformat().replace("+00:00", "Z"),
            "updated": created_process.updated.isoformat().replace("+00:00", "Z"),
            "finished": None,
            "activity_state": "disabled",
            "chunks": 1,
            "element": None,
            "element_name_contains": None,
            "element_type": None,
            "farm": None,
            "files": [],
            "folder_type": None,
            "load_children": False,
            "ml_class_id": None,
            "template_id": None,
            "use_cache": False,
            "tasks": [
                {
                    "depth": 0,
                    "finished": None,
                    "id": str(created_task.id),
                    "original_task_id": None,
                    "parents": [],
                    "requires_gpu": False,
                    "run": 0,
                    "shm_size": None,
                    "slug": "pdf_export",
                    "started": None,
                    "state": "unscheduled",
                    "ttl": 3600,
                }
            ],
        })

    def test_non_default_configuration(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(24):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "configuration": {
                        "page_type": "double_page",
                        "order_by_name": True,
                        "example_int": 54,
                        "example_choice": "b",
                        "example_list": ["beurre", "sucre"],
                        "example_required": "something"
                    }
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_process = Process.objects.get(mode=ProcessMode.Export)
        created_run = created_process.worker_runs.get()

        self.assertDictEqual(created_run.configuration.configuration, {
            "example_choice": "b",
            "example_int": 54,
            "example_list": ["beurre", "sucre"],
            "example_list_2": ["prune", "abricot"],
            "example_required": "something",
            "export_id": str(self.complete_export.id),
            "order_by_name": True,
            "page_type": "double_page"
        })

    @patch("arkindex.project.triggers.export.local_export.delay")
    def test_create_export_process_new_sql_export(self, delay_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(26), self.captureOnCommitCallbacks(execute=True):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                    "configuration": {
                        "example_required": "something"
                    }
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_process = Process.objects.get(mode=ProcessMode.Export)
        created_run = created_process.worker_runs.get()
        new_export = self.corpus.exports.order_by("-created").first()
        self.assertEqual(self.corpus.exports.count(), 2)
        self.assertNotEqual(new_export.id, self.complete_export.id)

        self.assertDictEqual(created_run.configuration.configuration, {
            "example_choice": "a",
            "example_int": 2,
            "example_list": ["pomme", "poire"],
            "example_list_2": ["prune", "abricot"],
            "example_required": "something",
            "export_id": str(new_export.id),
            "order_by_name": False,
            "page_type": "page"
        })

        # The new SQLite export has been started
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_export=new_export,
            user_id=self.user.id,
            description="Export of corpus Unit Tests",
        ))

    @patch("arkindex.project.triggers.export.local_export.delay")
    def test_create_export_process_new_sql_export_with_element(self, delay_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(30), self.captureOnCommitCallbacks(execute=True):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                    "configuration": {
                        "example_required": "something"
                    },
                    "element_id": str(self.corpus_element.id)
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_process = Process.objects.get(mode=ProcessMode.Export)
        created_run = created_process.worker_runs.get()
        new_export = self.corpus.exports.order_by("-created").first()
        self.assertEqual(self.corpus.exports.count(), 2)
        self.assertNotEqual(new_export.id, self.complete_export.id)
        self.assertEqual(created_process.element_id, self.corpus_element.id)

        self.assertDictEqual(created_run.configuration.configuration, {
            "example_choice": "a",
            "example_int": 2,
            "example_list": ["pomme", "poire"],
            "example_list_2": ["prune", "abricot"],
            "example_required": "something",
            "export_id": str(new_export.id),
            "order_by_name": False,
            "page_type": "page"
        })

        # The new SQLite export has been started
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_export=new_export,
            user_id=self.user.id,
            description="Export of corpus Unit Tests",
        ))

    @patch("arkindex.project.triggers.export.local_export.delay")
    def test_create_export_process_new_sql_export_with_selection(self, delay_mock):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.corpus_element)
        self.user.selected_elements.add(self.other_element)

        with self.assertNumQueries(29), self.captureOnCommitCallbacks(execute=True):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": None,
                    "configuration": {
                        "example_required": "something"
                    },
                    "selection": True
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_process = Process.objects.get(mode=ProcessMode.Export)
        created_run = created_process.worker_runs.get()
        new_export = self.corpus.exports.order_by("-created").first()
        self.assertEqual(self.corpus.exports.count(), 2)
        self.assertNotEqual(new_export.id, self.complete_export.id)
        self.assertEqual(list(created_process.elements.all()), [self.corpus_element])

        self.assertDictEqual(created_run.configuration.configuration, {
            "example_choice": "a",
            "example_int": 2,
            "example_list": ["pomme", "poire"],
            "example_list_2": ["prune", "abricot"],
            "example_required": "something",
            "export_id": str(new_export.id),
            "order_by_name": False,
            "page_type": "page"
        })

        # The new SQLite export has been started
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_export=new_export,
            user_id=self.user.id,
            description="Export of corpus Unit Tests",
        ))

    def test_farm(self):
        farm = Farm.objects.get()
        self.client.force_login(self.user)

        with self.assertNumQueries(25):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "farm_id": str(farm.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process = Process.objects.get(id=response.json()["id"])
        self.assertEqual(process.farm, farm)

    @patch("arkindex.process.serializers.imports.get_default_farm")
    def test_default_farm(self, get_default_farm_mock):
        farm = Farm.objects.get()
        get_default_farm_mock.return_value = farm
        self.client.force_login(self.user)

        with self.assertNumQueries(24):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process = Process.objects.get(id=response.json()["id"])
        self.assertEqual(process.farm, farm)

    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    def test_farm_unavailable(self, is_available_mock):
        farm = Farm.objects.get()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "farm_id": farm.id,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })
        self.assertFalse(self.corpus.processes.filter(mode=ProcessMode.Export).exists())

    def test_unknown_farm_id(self):
        self.client.force_login(self.user)

        farm_id = str(uuid4())
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:export-process", kwargs={"corpus_id": str(self.corpus.id)}),
                {
                    "format": "pdf",
                    "export_id": str(self.complete_export.id),
                    "farm_id": farm_id,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "farm_id": [f'Invalid pk "{farm_id}" - object does not exist.'],
        })
        self.assertFalse(self.corpus.processes.filter(mode=ProcessMode.Export).exists())
