from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse

from arkindex.process.models import ArkindexFeature, ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import Model, ModelVersionState


class TestProcessRun(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.init_worker_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        cls.import_worker_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        cls.version_with_model = WorkerVersion.objects.get(worker__slug="generic")
        cls.model = Model.objects.create(name="My model")
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

    @override_settings(PONOS_DEFAULT_ENV={"ARKINDEX_API_TOKEN": "testToken"})
    @patch("arkindex.process.builder.token_default")
    def test_pdf_import_run(self, token_mock):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Files,
        )
        process.files.create(name="A PDF file", size=10e6, content_type="pdf", corpus=self.corpus)
        token_mock.return_value = "12345"

        with patch("arkindex.process.tasks.initialize_activity.delay"):
            process.run()

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        import_task = process.tasks.get(slug="import_files")
        self.assertEqual(import_task.env, {
            "ARKINDEX_API_TOKEN": "testToken",
            "ARKINDEX_PROCESS_ID": str(process.id),
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
            "ARKINDEX_TASK_TOKEN": "12345"
        })
        self.assertIsNone(import_task.command)
        self.assertEqual(import_task.image, self.import_worker_version.docker_image_iid)

    @override_settings(
        PONOS_DEFAULT_ENV={"ARKINDEX_API_TOKEN": "testToken"},
        PUBLIC_HOSTNAME="https://arkindex.localhost",
    )
    @patch("arkindex.ponos.models.base64.encodebytes")
    def test_worker_run_model_version_run(self, token_mock):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
        )
        token_mock.side_effect = [b"12345", b"78945"]
        init_run = process.worker_runs.create(version=self.init_worker_version)
        run = process.worker_runs.create(version=self.version_with_model, parents=[init_run.id])
        run.model_version = self.model_version
        run.save()
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            process.run()

        initialisation_task = process.tasks.get(slug="initialisation")
        self.assertEqual(initialisation_task.env, {
            "ARKINDEX_API_TOKEN": "testToken",
            "ARKINDEX_PROCESS_ID": str(process.id),
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_TASK_TOKEN": "12345",
            "ARKINDEX_WORKER_RUN_ID": str(init_run.id)
        })
        self.assertEqual(initialisation_task.command, self.init_worker_version.docker_command)
        self.assertEqual(initialisation_task.image, self.init_worker_version.docker_image_iid)

        worker_task = process.tasks.get(slug=run.task_slug)
        self.assertEqual(worker_task.env, {
            "ARKINDEX_API_TOKEN": "testToken",
            "ARKINDEX_PROCESS_ID": str(process.id),
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_TASK_TOKEN": "78945",
            "ARKINDEX_WORKER_RUN_ID": str(run.id),
            "TASK_ELEMENTS": "/data/initialisation/elements.json"
        })
        self.assertEqual(worker_task.image, run.version.docker_image_iid)
        self.assertEqual(worker_task.command, run.version.docker_command)
        self.assertEqual(worker_task.extra_files, {
            "model": "https://arkindex.localhost" + reverse("api:model-version-download", kwargs={"pk": run.model_version.id}) + f"?token={run.model_version.build_authentication_token_hash()}"
        })
        self.assertEqual(len(worker_task.parents.all()), 1)
        self.assertEqual(worker_task.parents.first().slug, "initialisation")
