from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.ponos.models import Farm
from arkindex.process.models import FeatureUsage, Process, ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCorpusWorkerRuns(FixtureAPITestCase):
    """
    Test corpus worker run endpoints and methods
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.create(
            farm=Farm.objects.first(),
            creator=cls.user,
            mode=ProcessMode.Workers,
            corpus=cls.corpus
        )
        cls.dla_worker_version = WorkerVersion.objects.get(worker__slug="dla")
        cls.reco_worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.run_1 = WorkerRun.objects.create(
            process=cls.process,
            version=cls.dla_worker_version,
            has_results=True
        )

        cls.private_corpus = Corpus.objects.create(name="private")
        page_type = cls.private_corpus.types.create(
            slug="page",
            display_name="Page"
        )
        Element.objects.bulk_create(
            Element(
                corpus=cls.private_corpus,
                name=f"page {i}",
                type=page_type
            )
            for i in range(1, 11)
        )
        cls.private_process = Process.objects.create(
            farm=Farm.objects.first(),
            creator=cls.user,
            mode=ProcessMode.Workers,
            corpus=cls.private_corpus
        )
        cls.private_process.elements.set(Element.objects.filter(corpus=cls.private_corpus, type=page_type))
        cls.private_process.save()
        cls.run_2 = WorkerRun.objects.create(
            process=cls.private_process,
            version=cls.reco_worker_version,
        )
        cls.run_3 = WorkerRun.objects.create(
            process=cls.private_process,
            version=cls.dla_worker_version,
            has_results=True
        )

        cls.local_process = Process.objects.get(mode=ProcessMode.Local, creator=cls.user)
        cls.local_process.element = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.private_corpus.types.create(
                slug="book",
                display_name="Book",
                folder=True,
            ),
            name="Book One",
        )
        cls.local_process.save()
        cls.local_worker_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.user_local_run = WorkerRun.objects.get(process__mode=ProcessMode.Local, process__creator=cls.user)
        cls.user_local_run.has_results = True
        cls.user_local_run.save()

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_list_requires_read_access(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:corpus-runs", kwargs={"pk": self.private_corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_list(self):
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-runs", kwargs={"pk": self.private_corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["results"], [
            {
                "configuration": None,
                "id": str(self.user_local_run.id),
                "model_version": None,
                "parents": [],
                "process": {
                    "activity_state": "disabled",
                    "chunks": 1,
                    "corpus": None,
                    "element_id": str(self.local_process.element.id),
                    "element_count": 0,
                    "id": str(self.local_process.id),
                    "mode": "local",
                    "name": None,
                    "state": "unscheduled",
                    "use_cache": False,
                },
                "summary": "Worker Custom worker @ version 1",
                "worker_version": {
                    "configuration": {
                        "custom": "value"
                    },
                    "created": self.local_worker_version.created.isoformat().replace("+00:00", "Z"),
                    "docker_image_iid": None,
                    "gpu_usage": "disabled",
                    "id": str(self.local_worker_version.id),
                    "model_usage": FeatureUsage.Disabled.value,
                    "revision_url": None,
                    "state": "created",
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "worker": {
                        "id": str(self.local_worker_version.worker.id),
                        "name": "Custom worker",
                        "slug": "custom",
                        "type": "custom",
                        "description": "",
                        "repository_url": None,
                        "archived": False,
                        "cost_1k_elements": "0.000",
                        "cost_cpu_hour": "0.000",
                        "cost_gpu_hour": "0.000",
                    }
                },
                "use_gpu": False,
            },
            {
                "id": str(self.run_3.id),
                "worker_version": {
                    "id": str(self.dla_worker_version.id),
                    "configuration": {"test": 42},
                    "docker_image_iid": self.dla_worker_version.docker_image_iid,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "created": self.dla_worker_version.created.isoformat().replace("+00:00", "Z"),
                    "state": "available",
                    "worker": {
                        "id": str(self.dla_worker_version.worker.id),
                        "name": "Document layout analyser",
                        "slug": "dla",
                        "type": "dla",
                        "description": "",
                        "repository_url": None,
                        "archived": False,
                        "cost_1k_elements": "0.000",
                        "cost_cpu_hour": "0.000",
                        "cost_gpu_hour": "0.000",
                    }
                },
                "parents": [],
                "model_version": None,
                "configuration": None,
                "process": {
                    "id": str(self.private_process.id),
                    "activity_state": "disabled",
                    "corpus": str(self.private_corpus.id),
                    "chunks": 1,
                    "element_id": None,
                    "element_count": 10,
                    "mode": "workers",
                    "name": None,
                    "state": "unscheduled",
                    "use_cache": False,
                },
                "use_gpu": False,
                "summary": "Worker Document layout analyser @ version 1",
            }
        ])
