from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerRetrieve(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_custom = Worker.objects.get(slug="custom")

    def test_requires_login(self):
        response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_custom.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.worker_custom.id),
            "repository_url": None,
            "name": "Custom worker",
            "description": "",
            "slug": "custom",
            "type": "custom",
            "archived": False,
            "cost_1k_elements": "0.000",
            "cost_cpu_hour": "0.000",
            "cost_gpu_hour": "0.000",
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_no_rights(self, filter_rights_mock):
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
        ]
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Worker matches the given query."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])
